from save_to_db.core.item import Item
from save_to_db.exceptions import MultipleModelsMatch
from save_to_db.utils.test_base import TestBase


class TestItemHooks(TestBase):
    """ This class contains tests that make sure that item hooks are called at
    proper times and arguments.
    """
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    def setup_models(self,
                     one_dict=None, two_dict=None,
                     one_conf=None, two_conf=None,
                     clear_registry=False):
        if clear_registry:
            self.item_cls_manager.clear()
        
        one_dict = one_dict if one_dict is not None else {}
        one_dict['model_cls'] = self.ModelGeneralOne
        if one_conf:
            one_dict.update(one_conf)
        self.ItemGeneralOne = type('ItemGeneralOne', (Item,), one_dict)

        two_dict = one_dict if two_dict is not None else {}
        two_dict['model_cls'] = self.ModelGeneralTwo
        if two_conf:
            two_dict.update(two_conf)
        self.ItemGeneralTwo = type('ItemGeneralTwo', (Item,), two_dict)
    
    
    def test_before_and_after_process(self):
        process_log = []
        def before_process(item):
            self.assertIsInstance(item['f_integer'], str)  # still a string
            process_log.append('BEFORE {}: {} ({})'.format(
                item.__class__.__name__, item['f_integer'],
                type(item['f_integer']).__name__))
        
        def after_process(item):
            self.assertIsInstance(item['f_integer'], int)  # an integer
            process_log.append('AFTER {}: {} ({})'.format(
                item.__class__.__name__, item['f_integer'],
                type(item['f_integer']).__name__))
    
        self.setup_models(one_dict={'before_process': before_process,
                                    'after_process': after_process},
                          two_dict={'before_process': before_process,
                                    'after_process': after_process})
        
        item_one = self.ItemGeneralOne(f_integer='1000')
        for i in range(2):
            item_one['two_x_x'].gen(f_integer=str(2000+i))
        
        item_one.process()
        
        expected_log =  ['BEFORE ItemGeneralOne: 1000 (str)',
                         'BEFORE ItemGeneralTwo: 2000 (str)',
                         'AFTER ItemGeneralTwo: 2000 (int)',
                         'BEFORE ItemGeneralTwo: 2001 (str)',
                         'AFTER ItemGeneralTwo: 2001 (int)',
                         'AFTER ItemGeneralOne: 1000 (int)']
        
        self.assertEqual(process_log, expected_log)
    
    
    def test_before_and_after_persist(self):
        update_log = []
        
        def before_model_update(item, model):
            update_log.append([
                'before_model_update',
                [item['f_integer'], type(item['f_integer']).__name__],
                model.f_integer,
            ])
        
        def after_model_save(item, model):
            update_log.append([
                'after_model_save',
                [item['f_integer'], type(item['f_integer']).__name__],
                model.f_integer,
            ])
        
        persister = self.persister
        
        self.setup_models(one_dict={'before_model_update': before_model_update,
                                    'after_model_save': after_model_save},
                          one_conf={'creators': ['f_float'],
                                    'getters': ['f_float']},
                          two_dict={'before_model_update': before_model_update,
                                    'after_model_save': after_model_save},
                          two_conf={'creators': ['f_float'],
                                    'getters': ['f_float']},)
        
        # initial item persisted, some items in a bulk omitted -----------------
        item_one = self.ItemGeneralOne(f_integer='1000', f_float='10.1')
        for i in range(3):
            item_two = item_one['two_x_x'].gen(f_integer=str(2000+i))
            if i % 2 == 0:  # some items will be omitted
                item_two['f_float'] = str(2100+i)
        
        persister.persist(item_one)
        
        expected = [
            ['before_model_update', [2000, 'int'], None],
            ['after_model_save', [2000, 'int'], 2000],
            # item with `f_integer='2001'` was not created
            ['before_model_update', [2002, 'int'], None],
            ['after_model_save', [2002, 'int'], 2002],
            ['before_model_update', [1000, 'int'], None],
            ['after_model_save', [1000, 'int'], 1000]
        ]
        
        sort_func = lambda x: (x[0], x[1][0], x[2],)
        expected.sort(key=sort_func)
        update_log.sort(key=sort_func)
        self.assertEqual(update_log, expected)
        
        # initial item omitted -------------------------------------------------
        update_log.clear()
        
        item_one = self.ItemGeneralOne(f_integer='10000')
        # 1 item updated, 1 item created
        for i in range(2):
            item_one['two_x_x'].gen(f_integer=str(2000+i), f_float=str(2100+i))
        
        persister.persist(item_one)
        

        expected = [
            ['before_model_update', [2000, 'int'], 2000],  # updated
            ['after_model_save', [2000, 'int'], 2000],
            ['before_model_update', [2001, 'int'], None],  # created
            ['after_model_save', [2001, 'int'], 2001],
        ]
        
        expected.sort(key=sort_func)
        update_log.sort(key=sort_func)
        
        self.assertTrue(update_log == expected)
    
    
    def test_item_hooks_order(self):
        log = []
        def before_process(*_args, **_kwargs):
            log.append('before_process')
        def after_process(*_args, **_kwargs):
            log.append('after_process')
        def before_model_update(*_args, **_kwargs):
            log.append('before_model_update')
        def after_model_save(*_args, **_kwargs):
            log.append('after_model_save')
            
        self.setup_models(one_dict={'before_process': before_process,
                                    'after_process': after_process,
                                    'before_model_update': before_model_update,
                                    'after_model_save': after_model_save})
        
        item_one = self.ItemGeneralOne(f_integer='1000')
        self.persister.persist(item_one)
        
        expected_log = ['before_process',
                        'after_process',
                        'before_model_update',
                        'after_model_save']
        self.assertTrue(log, expected_log)
    



from collections import defaultdict
from .merge_items import merge_items
from save_to_db.exceptions import MultipleItemsMatch


def complete_item_structure(item):
    """ Takes an item, completes reverse relationships inside it and
    return a dictionary like this:
    
        .. code-block:: Python
            
            {
                item_cls: [item_instance, ...],
                ...
            }
            
    Where keys are a subclasses of :py:class:`~.item.Item` and values are
    instances of that class.
    """
    structure = defaultdict(list)  # item class to a list of instances
    __add_flat_item_to_structure(item, structure)
     
    # merging items and checking collisions
    all_items = []
    for items in structure.values():
        all_items.extend(items)
    for item_cls in structure.keys():
        if item_cls.allow_merge_items:
            merge_items(item, structure[item_cls], all_items)
        else:
            __check_item_collisions(structure[item_cls])
     
    return structure


def __add_flat_item_to_structure(item, structure):
    
    if item.is_single_item():
        if item in structure[type(item)]:
            return
        structure[type(item)].append(item)
        
        for key, relation in item.relations.items():
            if key not in item or item[key] is None:
                continue
            
            __add_flat_item_to_structure(item[key], structure)
            
            # setting reverse relations
            reverse_key = relation['reverse_key']
            if reverse_key is None:
                continue
            
            that_item = item[key]
            is_one_to_x = relation['relation_type'].is_one_to_x()
 
            if that_item.is_single_item():
                if is_one_to_x:
                    that_item[relation['reverse_key']] = item
                else:
                    that_item[relation['reverse_key']].add(item)
            else:
                for inner_item in that_item.bulk:
                    if is_one_to_x:
                        inner_item[reverse_key] = item
                    else:
                        try:
                            inner_item[reverse_key].add(item)
                        except:
                            raise Exception(key, reverse_key, is_one_to_x)
            
    else:
        # bulk item can be added (using this function) only once at the
        # beginning
        for inner_item in item.bulk:
            __add_flat_item_to_structure(inner_item, structure)


def __check_item_collisions(item_list):
    if not item_list:
        return
    
    # `item_list` must contain items of the same class
    getter_groups = item_list[0].getters
    
    if not getter_groups:
        return
    
    for left_item in item_list:
        for right_item in item_list:
            if left_item is right_item:
                continue
             
            for getters in getter_groups:
                same_by_group = True
                for key in getters:
                    if key not in left_item.data or key not in right_item.data \
                            or left_item[key] is None or \
                            left_item[key] != right_item[key]:
                        same_by_group = False
                        break
      
                     
                if same_by_group:
                    raise MultipleItemsMatch(None, left_item, right_item,
                                             getters)


import pkgutil
import importlib
from distutils.core import setup

import os
import sys
import save_to_db
print(save_to_db.__file__)
sys.path.insert(0, os.path.dirname(save_to_db.__file__))


packages = []
def add_package(package_load_path):
    global packages
    packages.append(package_load_path)
    package = importlib.import_module(package_load_path)
    for _, modname, ispkg in pkgutil.iter_modules(package.__path__):
        if not ispkg:
            continue
        add_package('{}.{}'.format(package_load_path, modname))
add_package('save_to_db')

def readme():
    with open('README.rst') as fp:
        return fp.read()

DESCRIPTION = ('This library makes it easy to store data from any source '
               'into a database.')

setup(
    name = 'Save-to-DB',
    packages = packages,
    version = '0.6.1a',
    description = DESCRIPTION,
    long_description = readme(),
    author = 'Mikhail Aleksandrovich Makovenko',
    author_email = 'mma86@rambler.ru',
    url = 'https://bitbucket.org/mikhail-makovenko/save-to-db',
    download_url = \
        'https://bitbucket.org/mikhail-makovenko/save-to-db/get/0.6.1a.tar.gz',
    keywords = 'development scraping mining database',
    license = 'MIT',
    classifiers = [
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
)



from functools import partial, wraps
from itertools import chain

from save_to_db.adapters.utils import adapter_manager
from save_to_db.adapters.utils.relation_type import RelationType

from save_to_db.exceptions import (InvalidFieldName,
                                   RelatedItemNotFound,
                                   MultipleRelatedItemsFound,
                                   ItemAdapterNotFound,
                                   NonExistentRelationDefined,
                                   NonExistentFieldUsed,
                                   FieldsFromRelationNotAllowed,
                                   NorewriteRelationException,
                                   NorewriteKeyUsedTwice,
                                   DeleterSelectorsOrKeepersEmpty,
                                   DeleterXToManyRelationUsed,
                                   UnrefNonXToManyFieldUsed)
from .item_cls_manager import item_cls_manager
from .model_deleter import ModelDeleter


class ItemMetaclass(type):
    """ A class that is used as a metaclass for :py:class:`~.item.Item`.
    The :py:class:`~.item.Item` class has a configuration of
    how collected data must be transformed and persisted into a database.
    This metaclass processes and corrects item configuration.
    """
    
    def __new__(cls, name, bases, dct, no_setup=False):
        item_cls = super().__new__(cls, name, bases, dct)
        
        # `no_setup` variable is only used to prevent setting up the
        # :py:class:`~.item.Item` class that must be only used as base class.
        if not no_setup:
            cls.process_configuration(item_cls)
            if item_cls.metadata['collection_id'] == None:
                item_cls_manager.add(item_cls)
                
        return item_cls
    
    
    # this is needed for Python versions less then 3.6
    def __init__(cls, name, bases, dct, **_):
        return super().__init__(name, bases, dct)
    
    
    @staticmethod
    def get_default_configuration():
        return {
            'batch_size': None,
            'metadata': {
                'collection_id': None,
                'setup_completed': False,
                'autogenerated_item_cls': False,
                'model_deleter': None,
                'model_unrefs': {},
            },
            'model_cls': None,
            'defaults': {},
            'creators': None,
            'creators_autoconfig': None,
            'autoinject_creators': True,
            'getters': None,
            'getters_autoconfig': None,
            'nullables': set(),
            'remove_null_fields': None,
            'remove_null_fields_autoconfig': None,
            'fields': {},
            'relations': {},
            'conversions': {
                # boolean
                'boolean_true_strings': ('true', 'yes', 'on', '1', '+',),
                'boolean_false_strings': ('false', 'no', 'off', '0', '-',),
                # integer and float
                'decimal_separator': '.',
                # date and time
                'date_formats': ('%Y-%m-%d',),
                'time_formats': ('%H:%M:%S',),
                'datetime_formats': ('%Y-%m-%d %H:%M:%S',),
            },
            'allow_multi_update': False,
            'allow_merge_items': False,
            'aliases': {},
            'update_only_mode': False,
            'get_only_mode': False,
            'norewrite_fields': {},
            'fast_insert': False,
            
            'deleter_selectors': None,
            'deleter_keepers': None,
            'deleter_execute_on_persist': False,
            
            'unref_x_to_many': {},
        }
    
    @staticmethod
    def get_default_relation_attrs():
        return {
            'item_cls': None,
            'relation_type': None,
            'replace_x_to_many': False,
            'reverse_key': None,
        }
    
    @classmethod
    def process_configuration(cls, item_cls):
        #--- making sure all configuration variables are present ---
        default_configuration = cls.get_default_configuration()
        for attr_name, default_value in default_configuration.items():
            if not hasattr(item_cls, attr_name):
                setattr(item_cls, attr_name, default_value)
                
        #--- correcting metadata -----------------------------------------------
        for key, value in default_configuration['metadata'].items():
            if key not in item_cls.metadata:
                item_cls.metadata[key] = value
        
        #--- correcting configuration structure --------------------------------
        
        # conversions
        conversions = default_configuration['conversions']
        conversions.update(item_cls.conversions)
        
        # fixing conversions for date and time
        for key in ('date_formats', 'time_formats', 'datetime_formats',):
            if isinstance(conversions[key], str):
                conversions[key] = (conversions[key],)
        
        item_cls.conversions = conversions
        
        # getters and creators 
        for groups_key in ('getters', 'creators'):
            groups = getattr(item_cls, groups_key)
            if groups is None:
                continue  # to be autoconfigured
            for i in range(len(groups)):
                if isinstance(groups[i], str):
                    groups[i] = {groups[i]}
                if not isinstance(groups[i], set):
                    groups[i] = set(groups[i])
        
        # nullables
        if not isinstance(item_cls.nullables, set):
            item_cls.nullables = set(item_cls.nullables)
        
        # remove_null_fields
        if item_cls.remove_null_fields is not None and \
                not isinstance(item_cls.remove_null_fields, set):
            item_cls.remove_null_fields = set(item_cls.remove_null_fields)
        
        # relations
        relations = item_cls.relations
        default_relation_attrs = cls.get_default_relation_attrs()
        for key in relations:
            if not isinstance(relations[key], dict):
                relations[key] = {
                    'item_cls': relations[key],
                }
            
            for k, v in default_relation_attrs.items():
                relations[key].setdefault(k, v)
        
        # unref_x_to_many
        unref_x_to_many = item_cls.unref_x_to_many
        for key in unref_x_to_many:
            if isinstance(unref_x_to_many[key], (list, set, tuple)):
                unref_x_to_many[key] = {
                    'selectors': unref_x_to_many[key],
                }
            if 'keepers' not in unref_x_to_many[key]:
                unref_x_to_many[key]['keepers'] = None
        
        #--- complete_setup ----------------------------------------------------
        wrap = wraps(cls.complete_setup)
        item_cls.complete_setup = wrap(partial(cls.complete_setup, item_cls))
            
    
    @classmethod
    def complete_setup(cls, item_cls):
        """ This method validates manual configuration of an
        :py:class:`~.item.Item` and automatically completes configuration based
        on available data.
        
        .. note::
            All :py:class:`~.item.Item` classes get this method wrapped in
            :py:class:`functools.partial` that already passes the `item_cls`
            value.
        """
        if item_cls.metadata['setup_completed']:
            return
        
        #--- relation paths ----------------------------------------------------
        for relation in item_cls.relations.values():
            if isinstance(relation['item_cls'], str):
                # first looking in the item's module
                item_cls_list = item_cls_manager.get_by_path(
                    relation['item_cls'],
                    relative_to=item_cls.__module__)
                # if not found, looking everywhere
                if not item_cls_list:
                    item_cls_list = item_cls_manager.get_by_path(
                        relation['item_cls'])
                    
                # exceptions
                if not item_cls_list:
                    raise RelatedItemNotFound(relation['item_cls'])
                if len(item_cls_list) != 1:
                    raise MultipleRelatedItemsFound(relation['item_cls'])
                relation['item_cls'] = item_cls_list[0]
        
        model_cls = item_cls.model_cls
        if model_cls is None:
            item_cls.metadata['setup_completed'] = True
            return  # no model_cls to fully complete setup
        
        #--- setup with adapter usage ------------------------------------------
        adapter_cls = adapter_manager.get_adapter_cls(model_cls)
        if not adapter_cls:
            raise ItemAdapterNotFound(item_cls)
        
        # fields
        item_cls.fields = {}
        for fname, ftype in adapter_cls.iter_fields(model_cls):
            if '__' in fname:
                raise InvalidFieldName(fname)
            item_cls.fields[fname] = ftype
        
        # relations
        original_relation_keys_to_copy = {'replace_x_to_many',}
        
        relations = {}
        for key, other_model_cls, relation_type, reverse_key in \
                adapter_cls.iter_relations(model_cls):
            other_item_cls_list = item_cls_manager.get_by_model_cls(
                other_model_cls, collection_id=item_cls.get_collection_id())
            # exceptions
            if not other_item_cls_list:
                raise RelatedItemNotFound(model_cls, key, other_model_cls)
            if len(other_item_cls_list) != 1:
                raise MultipleRelatedItemsFound(model_cls,
                                                other_item_cls_list)
            
            relations[key] = cls.get_default_relation_attrs()
            relations[key].update({
                'item_cls': other_item_cls_list[0],
                'relation_type': relation_type,
                'reverse_key': reverse_key,
            })
        
        for key in item_cls.relations.keys():
            if key not in relations:
                raise NonExistentRelationDefined(item_cls, 'relations', key)
            for subkey in original_relation_keys_to_copy:
                if subkey in item_cls.relations[key]:
                    relations[key][subkey] = \
                        item_cls.relations[key][subkey]
        
        # (to keep reference to the same dictionary)
        item_cls.relations.clear()
        item_cls.relations.update(relations)
        
        # norewrite_fields -----------------------------------------------------
        def fix_norewrite_field(item_cls, key, value):
            if key not in item_cls.norewrite_fields:
                item_cls.norewrite_fields[key] = value
            
            if key in item_cls.relations:
                other_item_cls = item_cls.relations[key]['item_cls']
                reverse_key = item_cls.relations[key]['reverse_key']
                if reverse_key:
                    if reverse_key not in other_item_cls.norewrite_fields:
                        other_item_cls.norewrite_fields[reverse_key] = value
                    elif not (value is
                              other_item_cls.norewrite_fields[reverse_key]):
                        raise NorewriteRelationException(
                            'norewrite_fields for relations must be the same '
                            'on both sides')
        
        # untuple norewrite fields
        def untuple(tupled_value):
            for key_in_tuple in tupled_value:
                if isinstance(key_in_tuple, tuple):
                    yield from untuple(key_in_tuple)
                else:
                    yield key_in_tuple
            
        for key in tuple(item_cls.norewrite_fields):
            if isinstance(key, tuple):
                norewrite_value = item_cls.norewrite_fields[key]
                del item_cls.norewrite_fields[key]
                for key_in_tuple in untuple(key):
                    if key_in_tuple in item_cls.norewrite_fields:
                        raise NorewriteKeyUsedTwice(key_in_tuple)
                    item_cls.norewrite_fields[key_in_tuple] = norewrite_value
        
        # regular fileds
        norewrite_selectors = {}
        for key in tuple(item_cls.norewrite_fields):
            norewrite_value = item_cls.norewrite_fields[key]
            if isinstance(key, (bool, RelationType)):  # selector field
                norewrite_selectors[key] = norewrite_value
                continue
            fix_norewrite_field(item_cls, key, norewrite_value)
        
        # selector fields
        selector_keys = list(norewrite_selectors.keys())
        type_to_key = {
            RelationType: 0,
            bool: 1,
        }
        selector_keys.sort(key=lambda t: type_to_key[type(t)])
        for selector in selector_keys:
            norewrite_value = item_cls.norewrite_fields[selector]
            del item_cls.norewrite_fields[selector]
            
            if isinstance(selector, RelationType):
                for key, relation in item_cls.relations.items():
                    if key in item_cls.norewrite_fields:
                        continue
                    if relation['relation_type'] is selector:
                        fix_norewrite_field(item_cls, key, norewrite_value)
                continue
            
            # boolean is the last in the list, it'll set all fields, including
            # relations
            if isinstance(selector, bool):
                for key in chain(item_cls.fields, item_cls.relations):
                    if key in item_cls.norewrite_fields:
                        continue
                    fix_norewrite_field(item_cls, key, norewrite_value)
                continue
        
        # creators, getters, nullables, and other field validation -------------
        def basic_field_validation(item_cls, field_names, groups_key):
            for field_name in field_names:
                
                if '__' in field_name:
                    raise FieldsFromRelationNotAllowed(item_cls, groups_key,
                                                      field_name)
                
                # NonExistentFieldUsed
                if field_name not in item_cls.fields and \
                        field_name not in item_cls.relations:
                    raise NonExistentFieldUsed(item_cls, groups_key,
                                               field_name)
                
                # Default relations
                if groups_key == 'defaults' and \
                        field_name in item_cls.relations:
                    default_value = item_cls.defaults[field_name]
                    if not isinstance(default_value, list):
                        item_cls.defaults[field_name] = default_value
                    else:
                        bulk_item_cls = \
                            item_cls.relations[field_name]['item_cls']
                        default_bulk = bulk_item_cls.Bulk()
                        default_bulk.add(*default_value)
                        item_cls.defaults[field_name] = default_bulk
                            
        
        for groups_key in ('defaults', 'getters', 'creators', 'nullables',
                           'remove_null_fields',
                           'norewrite_fields',
                           'deleter_selectors', 'deleter_keepers',
                           'unref_x_to_many',):
            groups = getattr(item_cls, groups_key)
            if groups is None:
                continue  # to be autoconfigured
            
            if groups_key == 'nullables':  # nullables is a single group
                groups = [groups]
            elif groups_key == 'defaults':  # defaults is a dictionary
                groups = [set(groups)]
            elif groups_key == 'norewrite_fields':
                groups = [set(groups)]
            elif groups_key == 'remove_null_fields':
                groups = [set(groups)]
            elif groups_key in ['deleter_selectors',
                                'deleter_keepers']:
                groups = [set(groups)]
            elif groups_key == 'unref_x_to_many':
                groups = [set(groups)]
                
            basic_field_validation(item_cls, chain(*groups), groups_key)
        
        # creators and getters automatic configuration -------------------------
        if item_cls.creators_autoconfig is True or \
                (item_cls.creators_autoconfig is None and
                 item_cls.creators is None):
            item_cls.creators = item_cls.creators or []
            not_nulls = set(adapter_cls.iter_required_fields(model_cls))
            
            if not_nulls:
                if not item_cls.creators:
                    # now `item_cls.creators` cannot be empty
                    item_cls.creators.append(not_nulls)
                    
                if item_cls.autoinject_creators:
                    for index, value in enumerate(item_cls.creators):
                        item_cls.creators[index] = value.union(
                            not_nulls)
                elif not_nulls not in item_cls.creators:
                    item_cls.creators.append(not_nulls)
                        
            
        if item_cls.getters_autoconfig is True or \
                (item_cls.getters_autoconfig is None and
                 item_cls.getters is None):
            item_cls.getters = item_cls.getters or []
            uniques = adapter_cls.iter_unique_field_combinations(model_cls)
            for unique_set in uniques:
                unique_set = set(unique_set)
                if unique_set not in item_cls.getters:
                    item_cls.getters.append(unique_set)
        
        # remove_null_fields automatic configuration ---------------------------
        if item_cls.remove_null_fields_autoconfig is True or \
                (item_cls.remove_null_fields is None and
                 item_cls.remove_null_fields_autoconfig is None):
            item_cls.remove_null_fields = item_cls.remove_null_fields or set()
            not_nulls = set(adapter_cls.iter_required_fields(model_cls))
            item_cls.remove_null_fields = \
                item_cls.remove_null_fields.union(not_nulls)
        else:
            item_cls.remove_null_fields = item_cls.remove_null_fields or set()
        
        # completing setup for related classes ---------------------------------
        item_cls.metadata['setup_completed'] = True
        for relation in item_cls.relations.values():
            relation['item_cls'].complete_setup()
        item_cls.metadata['setup_completed'] = False
        
        # setting up model deleters --------------------------------------------
        # validating `unref_x_to_many`
        for fkey in item_cls.unref_x_to_many:
            if fkey not in item_cls.relations or \
                    item_cls.relations[fkey]['relation_type'].is_x_to_one():
                raise UnrefNonXToManyFieldUsed(fkey)
            
            other_cls = item_cls.relations[fkey]['item_cls']
            fields = list(chain(
                item_cls.unref_x_to_many[fkey].get('selectors') or [],
                item_cls.unref_x_to_many[fkey].get('keepers') or []))
            
            basic_field_validation(other_cls, fields, None)
            
        
        # generating model deleters
        def create_model_deleter(item_cls, selectors, keepers):
            model_cls = item_cls.model_cls
            
            if keepers is None:
                keepers = set(adapter_cls.get_primary_key_names(model_cls))
            
            if not selectors or not keepers:
                raise DeleterSelectorsOrKeepersEmpty(
                    'Selectors and keepers cannot be empty')
            
            for field_name in chain(selectors, keepers):
                if field_name in item_cls.relations:
                    relation = item_cls.relations[field_name]
                    if relation['relation_type'].is_x_to_many():
                        raise DeleterXToManyRelationUsed(field_name)
                    
            return ModelDeleter(model_cls, selectors, keepers)
        
        # model_deleter 
        if item_cls.deleter_selectors is not None:
            item_cls.metadata['model_deleter'] = create_model_deleter(
                item_cls,
                item_cls.deleter_selectors,
                item_cls.deleter_keepers)
        
        # model_unrefs
        for fkey in item_cls.unref_x_to_many:
            other_cls = item_cls.relations[fkey]['item_cls']
            item_cls.metadata['model_unrefs'][fkey] = create_model_deleter(
                other_cls,
                item_cls.unref_x_to_many[fkey]['selectors'],
                item_cls.unref_x_to_many[fkey]['keepers'])
        
        # testing aliases ------------------------------------------------------
        item_cls.metadata['setup_completed'] = True
        try:
            for alias in item_cls.aliases.values():
                item_cls._get_real_keys(alias)
        except:
            item_cls.metadata['setup_completed'] = False
            raise

        item_cls.metadata['setup_completed'] = True


import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint

from ..config import Base



class __CommonColumns(object):
    id = sa.Column(sa.Integer, primary_key=True)
    
    f_binary = sa.Column(sa.LargeBinary, nullable=True)
    f_boolean = sa.Column(sa.Boolean, nullable=True)
    f_string = sa.Column(sa.String, nullable=True)
    f_text = sa.Column(sa.Text, nullable=True)
    f_integer = sa.Column(sa.Integer, nullable=True)
    f_float = sa.Column(sa.Float, nullable=True)
    f_date = sa.Column(sa.Date, nullable=True)
    f_time = sa.Column(sa.Time, nullable=True)
    f_datetime = sa.Column(sa.DateTime, nullable=True)
    
    __mapper_args__ = {
        'order_by': 'id',
    }



class ModelMergeOne(__CommonColumns, Base):
    __tablename__ = 'model_merge_one'
    
    two_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_merge_two.id', ondelete='CASCADE'),
        unique=True)
    two_1_1 = relationship(
        'ModelMergeTwo',
        back_populates='one_1_1',
        foreign_keys=[two_1_1_id])
    
    two_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_merge_two.id', ondelete='CASCADE'))
    two_x_1 = relationship(
        'ModelMergeTwo',
        back_populates='one_1_x',
        foreign_keys=[two_x_1_id])


class ModelMergeTwo(__CommonColumns, Base):
    __tablename__ = 'model_merge_two'
    
    one_1_1 = relationship(
        'ModelMergeOne',
        back_populates='two_1_1',
        foreign_keys='[ModelMergeOne.two_1_1_id,]',
        uselist=False)
    one_1_x = relationship(
        'ModelMergeOne',
        back_populates='two_x_1',
        foreign_keys='[ModelMergeOne.two_x_1_id,]',)
    
    three_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_merge_three.id', ondelete='CASCADE'),
        unique=True)
    three_1_1 = relationship(
        'ModelMergeThree',
        back_populates='two_1_1',
        foreign_keys=[three_1_1_id])
    
    three_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_merge_three.id', ondelete='CASCADE'))
    three_x_1 = relationship(
        'ModelMergeThree',
        back_populates='two_1_x',
        foreign_keys=[three_x_1_id])


class ModelMergeThree(__CommonColumns, Base):
    __tablename__ = 'model_merge_three'
    
    two_1_1 = relationship(
        'ModelMergeTwo',
        back_populates='three_1_1',
        foreign_keys='[ModelMergeTwo.three_1_1_id,]',
        uselist=False)
    two_1_x = relationship(
        'ModelMergeTwo',
        back_populates='three_x_1',
        foreign_keys='[ModelMergeTwo.three_x_1_id,]',)


class ModelMergeOneX(__CommonColumns, Base):
    __tablename__ = 'model_merge_one_x'

    two_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_merge_two_x.id', ondelete='CASCADE'))
    two_x_1 = relationship(
        'ModelMergeTwoX',
        back_populates='one_1_x',
        foreign_keys=[two_x_1_id])
    
    __table_args__ = (
        UniqueConstraint('two_x_1_id', 'f_integer', name='unique_1'),
    )


class ModelMergeTwoX(__CommonColumns, Base):
    __tablename__ = 'model_merge_two_x'

    one_1_x = relationship(
        'ModelMergeOneX',
        back_populates='two_x_1',
        foreign_keys='[ModelMergeOneX.two_x_1_id,]',)
    
    three_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_merge_three_x.id', ondelete='CASCADE'))
    three_x_1 = relationship(
        'ModelMergeThreeX',
        back_populates='two_1_x',
        foreign_keys=[three_x_1_id])
    
    __table_args__ = (
        UniqueConstraint('three_x_1_id', 'f_integer', name='unique_1'),
    )


class ModelMergeThreeX(__CommonColumns, Base):
    __tablename__ = 'model_merge_three_x'

    two_1_x = relationship(
        'ModelMergeTwoX',
        back_populates='three_x_1',
        foreign_keys='[ModelMergeTwoX.three_x_1_id,]',)
    


from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestReplace(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    ModelManyRefsOne = None
    ModelManyRefsTwo = None
    
    
    def test_replace_x_to_many(self):
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_string'}]
            getters = [{'f_string'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_string'}]
            getters = [{'f_string'}]
            relations = {
                'one_1_x': {
                    'replace_x_to_many': True,
                }
            }
        
        persister = self.persister
        
        # replace_x_to_many = False
        item_one = ItemGeneralOne(f_string='1')
        item_one['two_1_x'].gen(f_string='1: 2->1')
        item_one['two_1_x'].gen(f_string='2: 2->1')
        persister.persist(item_one)
        item_one = ItemGeneralOne(f_string='1')  # item recreated
        item_one['two_1_x'].gen(f_string='3: 2->1')
        item_one['two_1_x'].gen(f_string='4: 2->1')
        
        _, model_lists = persister.persist(item_one)
        self.assertEqual(len(model_lists), 1)
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        model = models[0]
        two_1_x = self.get_related_x_to_many(model, 'two_1_x')
        self.assertEqual(len(two_1_x), 4)
        two_1_x.sort(key=lambda m: m.f_string)
        for i in range(4):
            self.assertEqual(two_1_x[i].f_string, '{}: 2->1'.format(i+1))
        
        # replace_x_to_many = True
        item_two = ItemGeneralTwo(f_string='2')
        item_two['one_1_x'].gen(f_string='1: 1->2')
        item_two['one_1_x'].gen(f_string='2: 1->2')
        persister.persist(item_two)
        item_two = ItemGeneralTwo(f_string='2')  # item recreated
        item_two['one_1_x'].gen(f_string='3: 1->2')
        item_two['one_1_x'].gen(f_string='4: 1->2')
        
        _, model_lists = persister.persist(item_two)
        self.assertEqual(len(model_lists), 1)
        models = model_lists[0]
        self.assertEqual(len(models), 1)
        model = models[0]
        one_1_x = self.get_related_x_to_many(model, 'one_1_x')
        self.assertEqual(len(one_1_x), 2)
        
        one_1_x.sort(key=lambda m: m.f_string)
        for i in range(2):
            # `i+3` because first two items replaced
            self.assertEqual(one_1_x[i].f_string, '{}: 1->2'.format(i+3))
    
    
    def __test_replace_x_to_many_additional_refrence(self,
                                                     one_1_x_a='one_1_x_a',
                                                     two_1_x_b='two_1_x_b'):

        class ItemManyRefsOne(Item):
            model_cls = self.ModelManyRefsOne
            creators = [{'f_string'}]
            getters = [{'f_string'}]
            relations = {
                'two_1_x_b': {
                    'replace_x_to_many': True,
                },
                'two_x_x_b': {
                    'replace_x_to_many': True,
                }
            }
        
        class ItemManyRefsTwo(Item):
            model_cls = self.ModelManyRefsTwo
            creators = [{'f_string'}]
            getters = [{'f_string'}]
            relations = {
                'one_1_x_a': {
                    'replace_x_to_many': True,
                },
                'one_x_x_a': {
                    'replace_x_to_many': True,
                }
            }
        
        persister = self.persister
        
        # --- creating ---
        item_two = ItemManyRefsTwo(f_string='Item 2.1')
        item_one = item_two[one_1_x_a].gen(f_string='Item 1.1')
        item_one[two_1_x_b].add(item_two)
        
        persister.persist(item_two)
        
        models_one = self.get_all_models(self.ModelManyRefsOne)
        models_two = self.get_all_models(self.ModelManyRefsTwo)
        self.assertEqual(len(models_one), 1)
        self.assertEqual(len(models_two), 1)
        self.assertEqual(models_one[0].f_string, 'Item 1.1')
        self.assertEqual(models_two[0].f_string, 'Item 2.1')
        
        x_models_two = self.get_related_x_to_many(models_one[0], two_1_x_b)
        x_models_one = self.get_related_x_to_many(models_two[0], one_1_x_a)
        self.assertEqual(len(x_models_one), 1)
        self.assertEqual(len(x_models_two), 1)
        
        self.assertEqual(x_models_one[0], models_one[0])
        self.assertEqual(x_models_two[0], models_two[0])
        
        # --- updating ---
        # item_two
        item_two = ItemManyRefsTwo(f_string='Item 2.1')
        item_two[one_1_x_a].gen(f_string='Item 1.2')
        # old item with old reference
        item_two['one_1_1_b'] = ItemManyRefsOne(f_string='Item 1.1')
        
        # item_one
        item_one = item_two['one_1_1_b']
        item_one[two_1_x_b].gen(f_string='Item 2.2')
        # old item with old reference
        item_one['two_1_1_a'] = item_two
        
        persister.persist(item_two)
        
        sort_func = lambda model: model.f_string
        
        models_one = self.get_all_models(self.ModelManyRefsOne)
        models_two = self.get_all_models(self.ModelManyRefsTwo)
        self.assertEqual(len(models_one), 2)
        self.assertEqual(len(models_two), 2)
        
        models_one.sort(key=sort_func)
        models_two.sort(key=sort_func)
        
        self.assertEqual(models_one[0].f_string, 'Item 1.1')
        self.assertEqual(models_one[1].f_string, 'Item 1.2')
        self.assertEqual(models_two[0].f_string, 'Item 2.1')
        self.assertEqual(models_two[1].f_string, 'Item 2.2')
        
        # updated relations must be without old references
        x_models_two = self.get_related_x_to_many(models_one[0], two_1_x_b)
        x_models_one = self.get_related_x_to_many(models_two[0], one_1_x_a)
        self.assertEqual(len(x_models_one), 1)
        self.assertEqual(len(x_models_two), 1)
        
        self.assertEqual(x_models_one[0], models_one[1])
        self.assertEqual(x_models_two[0], models_two[1])
        
        self.assertEqual(models_one[0].two_1_1_a, models_two[0])
        self.assertEqual(models_two[0].one_1_1_b, models_one[0])
    
    
    def test_replace_one_to_many_additional_refrence(self):
        self.__test_replace_x_to_many_additional_refrence(one_1_x_a='one_1_x_a',
                                                          two_1_x_b='two_1_x_b')
    
    
    def test_replace_many_to_many_additional_refrence(self):
        self.__test_replace_x_to_many_additional_refrence(one_1_x_a='one_x_x_a',
                                                          two_1_x_b='two_x_x_b')
    
    
    def test_one_to_one_replacement(self):

        class ItemManyRefsOne(Item):
            allow_merge_items = True
            model_cls = self.ModelManyRefsOne
            creators = [{'f_string'}]
            getters = [{'f_string'}]
        
        class ItemManyRefsTwo(Item):
            allow_merge_items = True
            model_cls = self.ModelManyRefsTwo
            creators = [{'f_string'}]
            getters = [{'f_string'}]
        
        persister = self.persister
        sort_func = lambda model: model.f_string
        
        # creating
        item_two = ItemManyRefsTwo(f_string='Item 2.1')
        item_two['one_1_1_a'] = ItemManyRefsOne(f_string='Item 1.1')
         
        persister.persist(item_two)
         
        models_one = self.get_all_models(self.ModelManyRefsOne)
        models_two = self.get_all_models(self.ModelManyRefsTwo)
        self.assertEqual(len(models_one), 1)
        self.assertEqual(len(models_two), 1)
        self.assertEqual(models_two[0].f_string, 'Item 2.1')
        self.assertEqual(models_one[0].f_string, 'Item 1.1')
         
        self.assertEqual(models_two[0].one_1_1_a, models_one[0])

        # updating without pulling old item
        item_two = ItemManyRefsTwo(f_string='Item 2.1')
        item_two['one_1_1_a'] = ItemManyRefsOne(f_string='Item 1.2')
        persister.persist(item_two)
        
        models_one = self.get_all_models(self.ModelManyRefsOne, sort_func)
        models_two = self.get_all_models(self.ModelManyRefsTwo, sort_func)
        self.assertEqual(len(models_one), 2)
        self.assertEqual(len(models_two), 1)
        self.assertEqual(models_one[0].f_string, 'Item 1.1')
        self.assertEqual(models_one[0].two_1_1_a, None)
        self.assertEqual(models_one[1].f_string, 'Item 1.2')
        self.assertEqual(models_one[1].two_1_1_a, models_two[0])
        self.assertEqual(models_two[0].f_string, 'Item 2.1')
        self.assertEqual(models_two[0].one_1_1_a, models_one[1])
    
    
    def test_one_to_one_additional_refrence(self):

        class ItemManyRefsOne(Item):
            allow_merge_items = True
            model_cls = self.ModelManyRefsOne
            creators = [{'f_string'}]
            getters = [{'f_string'}]
        
        class ItemManyRefsTwo(Item):
            allow_merge_items = True
            model_cls = self.ModelManyRefsTwo
            creators = [{'f_string'}]
            getters = [{'f_string'}]
        
        persister = self.persister
        sort_func = lambda model: model.f_string
        
        #  creating 
        item_two = ItemManyRefsTwo(f_string='Item 2.1')
        item_two['one_1_1_a'] = ItemManyRefsOne(f_string='Item 1.1')
        item_two['one_1_1_a']['two_1_1_b'] = \
            ItemManyRefsTwo(f_string='Item 2.2')
        
        persister.persist(item_two)
         
        models_one = self.get_all_models(self.ModelManyRefsOne, sort_func)
        models_two = self.get_all_models(self.ModelManyRefsTwo, sort_func)
        self.assertEqual(len(models_one), 1)
        self.assertEqual(len(models_two), 2)
        self.assertEqual(models_two[0].f_string, 'Item 2.1')
        self.assertEqual(models_one[0].f_string, 'Item 1.1')
        self.assertEqual(models_two[1].f_string, 'Item 2.2')
        
        self.assertEqual(models_one[0].two_1_1_a, models_two[0])
        self.assertEqual(models_one[0].two_1_1_b, models_two[1])
        
        #  updating with pulling old item 
        item_two = ItemManyRefsTwo(f_string='Item 2.1')
        item_two['one_1_1_a'] = ItemManyRefsOne(f_string='Item 1.2')
        item_one = item_two['one_x_x_a'].gen(f_string='Item 1.1')
        item_one['two_1_1_b'] = ItemManyRefsTwo(f_string='Item 2.3')
        # this item can potentially keep reference to item two
        item_one['two_x_x_b'].gen(f_string='Item 2.2')
          
        persister.persist(item_two)
        models_one = self.get_all_models(self.ModelManyRefsOne, sort_func)
        models_two = self.get_all_models(self.ModelManyRefsTwo, sort_func)
        self.assertEqual(len(models_one), 2)
        self.assertEqual(len(models_two), 3)
        self.assertEqual(models_two[0].f_string, 'Item 2.1')
        self.assertEqual(models_two[1].f_string, 'Item 2.2')
        self.assertEqual(models_two[2].f_string, 'Item 2.3')
        self.assertEqual(models_one[0].f_string, 'Item 1.1')
        self.assertEqual(models_one[1].f_string, 'Item 1.2')
        
        self.assertFalse(hasattr(models_one[0], 'two_1_1_a') and
                         models_one[0].two_1_1_a)
        self.assertEqual(models_one[0].two_1_1_b, models_two[2])
        
        self.assertEqual(models_two[0].one_1_1_a, models_one[1])
        self.assertFalse(hasattr(models_two[0], 'one_1_1_b') and
                         models_two[0].one_1_1_b)
        
        self.assertFalse(hasattr(models_two[1], 'one_1_1_a') and
                         models_two[1].one_1_1_a)
        self.assertFalse(hasattr(models_two[1], 'one_1_1_b') and
                         models_two[1].one_1_1_b)
        
        x_models_two = self.get_related_x_to_many(models_one[0], 'two_x_x_b')
        self.assertEqual(len(x_models_two), 1)
        self.assertEqual(x_models_two[0].f_string, 'Item 2.2')
        
        x_models_one = self.get_related_x_to_many(models_two[0], 'one_x_x_a')
        self.assertEqual(len(x_models_one), 1)
        self.assertEqual(x_models_one[0].f_string, 'Item 1.1')


from save_to_db.exceptions import CannotMergeModels, MultipleModelsMatch
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestResolveModel(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    def setUp(self):
        super().setUp()
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'f_text'}]
        self.ItemGeneralOne = ItemGeneralOne
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'f_text'}]
        self.ItemGeneralTwo = ItemGeneralTwo
        
        persister = self.persister
        
        bulk = ItemGeneralOne.Bulk()
        bulk.gen(f_integer='1', f_float='1.1', f_text='text-1')
        bulk.gen(f_integer='2', f_float='1.1', f_text='text-2')
        persister.persist(bulk)
        
        item = ItemGeneralOne(f_integer='2', f_text='text-1')
        persister.persist(item)
        
        with self.assertRaises(MultipleModelsMatch):
            item = ItemGeneralOne(f_integer='2', f_float='1.1', f_text='text-1')
            persister.persist(item)
            
        models = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models), 2)
    
    
    def __resolve_model(self, models):
        # sorting models to always get the same result after merging
        db_adapter = self.persister.db_adapter
        models = list(models)
        pk_list = list(db_adapter.get_primary_key_names(type(models[0])))
        pk_list.sort()
        get_model_key = lambda model: tuple(getattr(model, field)
                                            for field in pk_list)
        models.sort(key=get_model_key)
        
        result_model = db_adapter.merge_models(models)
        
        return result_model
    
    
    def test_merging_plain_fields(self):
        self.ItemGeneralOne.resolve_model = self.__resolve_model
        
        bulk = self.ItemGeneralOne.Bulk()
        bulk.gen(f_integer='1', f_string='string-1')
        bulk.gen(f_integer='2', f_boolean=True, f_date='2010-10-10')
        self.persister.persist(bulk)
        
        models = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models), 2)
        
        item = self.ItemGeneralOne(f_integer='2',
                                   f_float='1.1', f_text='text-1',
                                   f_date='2010-10-20',
                                   f_time='10:20:30')
        items, model_lists = self.persister.persist(item)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        models = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models), 1)  # second model was deleted
        
        model = models[0]
        # old values
        self.assertEqual(model.f_integer, 2)
        self.assertEqual(model.f_float, 1.1)
        self.assertEqual(model.f_text, 'text-1')
        # values from different models merged
        self.assertEqual(model.f_string, 'string-1')
        self.assertEqual(model.f_boolean, True)
        # values from item
        # (model value overwritten with item value)
        self.assertEqual(model.f_date.strftime('%Y-%m-%d'), '2010-10-20')
        # (new value from item)
        self.assertEqual(model.f_time.strftime('%H:%M:%S'), '10:20:30')
    
    
    def test_merging_x_to_many(self):
        persister = self.persister
        self.ItemGeneralOne.resolve_model = self.__resolve_model
        sort_func = lambda model: model.f_integer
        
        item_two_1 = self.ItemGeneralTwo(f_integer=1, f_string='two-1')
        item_two_2 = self.ItemGeneralTwo(f_integer=2, f_string='two-2')
        item_two_3 = self.ItemGeneralTwo(f_integer=3, f_string='two-3')
        item_two_4 = self.ItemGeneralTwo(f_integer=4, f_string='two-4')
        item_two_5 = self.ItemGeneralTwo(f_integer=5, f_string='two-5')
        item_two_6 = self.ItemGeneralTwo(f_integer=6, f_string='two-6')
        item_two_7 = self.ItemGeneralTwo(f_integer=7,
                                         f_string='two-7 (in item)')
        item_two_8 = self.ItemGeneralTwo(f_integer=8,
                                         f_string='two-8 (in item)')
        
        bulk = self.ItemGeneralOne.Bulk()
        item_one_1 = bulk.gen(f_integer='1')
        item_one_2 = bulk.gen(f_integer='2')
        
        # one-to-many
        item_one_1['two_1_x'].add(item_two_1, item_two_2)
        item_one_2['two_1_x'].add(item_two_3)
        
        # many-to-many
        item_one_1['two_x_x'].add(item_two_4, item_two_5)
        item_one_2['two_x_x'].add(item_two_5, item_two_6)
        
        persister.persist(bulk)
        
        models = self.get_all_models(self.ModelGeneralOne, sort_key=sort_func)
        self.assertEqual(len(models), 2)
        
        model_one_1, model_one_2 = models
        self.assertEqual(model_one_1.f_integer, 1)
        self.assertEqual(model_one_2.f_integer, 2)
        
        one_1_two_1_x = self.get_related_x_to_many(model_one_1, 'two_1_x',
                                                   sort_key=sort_func)
        one_1_two_x_x = self.get_related_x_to_many(model_one_1, 'two_x_x',
                                                   sort_key=sort_func)
        one_2_two_1_x = self.get_related_x_to_many(model_one_2, 'two_1_x',
                                                 sort_key=sort_func)
        one_2_two_x_x = self.get_related_x_to_many(model_one_2, 'two_x_x',
                                                   sort_key=sort_func)
        
        self.assertEqual(len(one_1_two_1_x), 2)
        self.assertEqual(one_1_two_1_x[0].f_integer, 1)
        self.assertEqual(one_1_two_1_x[1].f_integer, 2)
        
        self.assertEqual(len(one_2_two_1_x), 1)
        self.assertEqual(one_2_two_1_x[0].f_integer, 3)
        
        self.assertEqual(len(one_1_two_x_x), 2)
        self.assertEqual(one_1_two_x_x[0].f_integer, 4)
        self.assertEqual(one_1_two_x_x[1].f_integer, 5)
        
        self.assertEqual(len(one_2_two_x_x), 2)
        self.assertEqual(one_2_two_x_x[0].f_integer, 5)
        self.assertEqual(one_2_two_x_x[1].f_integer, 6)
        
        item = self.ItemGeneralOne(f_integer='2',
                                   f_float='1.1', f_text='text-1',
                                   two_1_x=[item_two_7],
                                   two_x_x=[item_two_8])
        items, model_lists = persister.persist(item)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        # now checking that all was merged correctly
        models = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models), 1)  # second model was deleted
        
        model = models[0]
        self.assertEqual(model.f_integer, 2)
        self.assertEqual(model.f_float, 1.1)
        self.assertEqual(model.f_text, 'text-1')
        two_1_x = self.get_related_x_to_many(model, 'two_1_x',
                                             sort_key=sort_func)
        two_x_x = self.get_related_x_to_many(model, 'two_x_x',
                                             sort_key=sort_func)
        self.assertEqual(len(two_1_x), 4)
        self.assertEqual(two_1_x[0].f_integer, 1)
        self.assertEqual(two_1_x[1].f_integer, 2)
        self.assertEqual(two_1_x[2].f_integer, 3)
        self.assertEqual(two_1_x[3].f_integer, 7)  # from item
        
        self.assertEqual(len(two_x_x), 4)
        self.assertEqual(two_x_x[0].f_integer, 4)
        self.assertEqual(two_x_x[1].f_integer, 5)
        self.assertEqual(two_x_x[2].f_integer, 6)
        self.assertEqual(two_x_x[3].f_integer, 8)  # from item
    

    def test_merging_x_to_one(self):
        self.ItemGeneralOne.resolve_model = self.__resolve_model
        sort_func = lambda model: model.f_integer
        
        item_two_1 = self.ItemGeneralTwo(f_integer=1, f_string='two-1')
        item_two_2 = self.ItemGeneralTwo(f_integer=2, f_string='two-2')
        
        bulk = self.ItemGeneralOne.Bulk()
        item_one_1 = bulk.gen(f_integer='1')
        item_one_2 = bulk.gen(f_integer='2')
        
        # same reference
        item_one_1['two_x_1'] = item_two_1
        item_one_2['two_x_1'] = item_two_1
        
        # one reference
        item_one_1['two_1_1'] = item_two_2
        
        self.persister.persist(bulk)
        models = self.get_all_models(self.ModelGeneralOne, sort_key=sort_func)
        self.assertEqual(len(models), 2)
        
        self.assertEqual(models[0].two_x_1.f_integer, 1)
        self.assertEqual(models[1].two_x_1.f_integer, 1)
        
        self.assertEqual(models[0].two_1_1.f_integer, 2)
        
        # merging models
        item = self.ItemGeneralOne(f_integer='2',
                                   f_float='1.1', f_text='text-1')
        items, model_lists = self.persister.persist(item)

        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        models = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models), 1)
        self.assertEqual(models[0].two_x_1.f_integer, 1)
        self.assertEqual(models[0].two_1_1.f_integer, 2)
        
    
    def test_merging_x_to_one_overwrite(self):
        persister = self.persister
        self.ItemGeneralOne.resolve_model = self.__resolve_model
        sort_func = lambda model: model.f_integer
        
        item_two_1 = self.ItemGeneralTwo(f_integer=1, f_string='two-1')
        item_two_2 = self.ItemGeneralTwo(f_integer=2, f_string='two-2')
        item_two_3 = self.ItemGeneralTwo(f_integer=3,
                                         f_string='two-3 (in item)')
        item_two_4 = self.ItemGeneralTwo(f_integer=4,
                                         f_string='two-4 (in item)')
        
        bulk = self.ItemGeneralOne.Bulk()
        item_one_1 = bulk.gen(f_integer='1')
        item_one_2 = bulk.gen(f_integer='2')
        
        # same reference
        item_one_1['two_x_1'] = item_two_1
        item_one_2['two_x_1'] = item_two_1
        
        # one reference
        item_one_1['two_1_1'] = item_two_2
        
        persister.persist(bulk)
        models = self.get_all_models(self.ModelGeneralOne, sort_key=sort_func)
        self.assertEqual(len(models), 2)
        
        self.assertEqual(models[0].two_x_1.f_integer, 1)
        self.assertEqual(models[1].two_x_1.f_integer, 1)
        
        self.assertEqual(models[0].two_1_1.f_integer, 2)
        
        # merging models
        item = self.ItemGeneralOne(f_integer='2',
                                   f_float='1.1', f_text='text-1',
                                   two_x_1=item_two_3, two_1_1=item_two_4)
        items, model_lists = persister.persist(item)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        models = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(models), 1)
        self.assertEqual(models[0].two_x_1.f_integer, 3)
        self.assertEqual(models[0].two_1_1.f_integer, 4)
        
        
    def test_merging_many_to_one_exception(self):
        persister = self.persister
        self.ItemGeneralOne.resolve_model = self.__resolve_model
        sort_func = lambda model: model.f_integer
        
        item_two_1 = self.ItemGeneralTwo(f_integer=1, f_string='two-1')
        item_two_2 = self.ItemGeneralTwo(f_integer=2, f_string='two-2')
        
        bulk = self.ItemGeneralOne.Bulk()
        bulk.gen(f_integer='1', two_x_1=item_two_1)
        bulk.gen(f_integer='2', two_x_1=item_two_2)
    
        persister.persist(bulk)
        
        models = self.get_all_models(self.ModelGeneralOne, sort_key=sort_func)
        self.assertEqual(len(models), 2)
        self.assertEqual(models[0].two_x_1.f_integer, 1)
        self.assertEqual(models[1].two_x_1.f_integer, 2)
        
        # merging models
        item = self.ItemGeneralOne(f_integer='2',
                                   f_float='1.1', f_text='text-1')
        
        with self.assertRaises(CannotMergeModels):
            persister.persist(item)
            
            
    def test_merging_one_to_one_exception(self):
        persister = self.persister
        self.ItemGeneralOne.resolve_model = self.__resolve_model
        sort_func = lambda model: model.f_integer
        
        item_two_1 = self.ItemGeneralTwo(f_integer=1, f_string='two-1')
        item_two_2 = self.ItemGeneralTwo(f_integer=2, f_string='two-2')
        
        bulk = self.ItemGeneralOne.Bulk()
        bulk.gen(f_integer='1', two_1_1=item_two_1)
        bulk.gen(f_integer='2', two_1_1=item_two_2)
    
        persister.persist(bulk)
        
        models = self.get_all_models(self.ModelGeneralOne, sort_key=sort_func)
        self.assertEqual(len(models), 2)
        self.assertEqual(models[0].two_1_1.f_integer, 1)
        self.assertEqual(models[1].two_1_1.f_integer, 2)
        
        # merging models
        item = self.ItemGeneralOne(f_integer='2',
                                   f_float='1.1', f_text='text-1')
        
        with self.assertRaises(CannotMergeModels):
            persister.persist(item)
    
    
    def test_allow_multi_update_true(self):
        persister = self.persister
        self.ItemGeneralOne.resolve_model = self.__resolve_model
        self.ItemGeneralOne.allow_multi_update = True
        sort_func = lambda model: model.f_integer
        
        item_two_1 = self.ItemGeneralTwo(f_integer=1, f_string='two-1')
        item_two_2 = self.ItemGeneralTwo(f_integer=2, f_string='two-2')
        
        bulk = self.ItemGeneralOne.Bulk()
        bulk.gen(f_integer='1', two_x_x=[item_two_1])
        bulk.gen(f_integer='2', two_x_x=[item_two_2])
    
        persister.persist(bulk)
        
        models = self.get_all_models(self.ModelGeneralOne, sort_key=sort_func)
        self.assertEqual(len(models), 2)
        
        self.assertEqual(models[0].f_boolean, None)
        self.assertEqual(models[1].f_boolean, None)
        
        two_x_x_1 = self.get_related_x_to_many(models[0], 'two_x_x')
        two_x_x_2 = self.get_related_x_to_many(models[1], 'two_x_x')
        
        self.assertEqual(len(two_x_x_1), 1)
        self.assertEqual(len(two_x_x_2), 1)
        
        self.assertEqual(two_x_x_1[0].f_integer, 1)
        self.assertEqual(two_x_x_2[0].f_integer, 2)
        
        # no merging models
        item = self.ItemGeneralOne(f_integer='2',
                                   f_float='1.1', f_text='text-1',
                                   f_boolean=True)
        
        persister.persist(item)
        
        models = self.get_all_models(self.ModelGeneralOne, sort_key=sort_func)
        self.assertEqual(len(models), 2)  # still two items
        
        # `f_boolean` updated for both models
        self.assertEqual(models[0].f_boolean, True)
        self.assertEqual(models[1].f_boolean, True)
        
        # but the rest stays the same
        two_x_x_1 = self.get_related_x_to_many(models[0], 'two_x_x')
        two_x_x_2 = self.get_related_x_to_many(models[1], 'two_x_x')
        
        self.assertEqual(len(two_x_x_1), 1)
        self.assertEqual(len(two_x_x_2), 1)
        
        self.assertEqual(two_x_x_1[0].f_integer, 1)
        self.assertEqual(two_x_x_2[0].f_integer, 2)
        
        


from save_to_db.exceptions.item_cls_collection import CollectionDoesNotExist


class ItemClsCollection(object):
    """ Class for managing collections of
    :py:class:`~save_to_db.core.item_base.ItemBase` sub-classes.
    
    :cvar collection:
        Set of all item classes in the collection.
    """
    collection_classes = {}
    
    def __init__(self, collection_id):
        self.collection = set()
        self.collection_id = collection_id
        self.collection_classes[collection_id] = self
    
    def add(self, *item_classes):
        """ Adds item classes to the collection.
        
        :param \*item_classes: List of item classes to add.
        """
        self.collection.update(item_classes)
    
    def remove(self, *item_classes):
        """ Removes item classes from the collection.
        
        :param \*item_classes: List of item classes to remove.
        """
        self.collection -= set(item_classes)
    
    def clear(self):
        """ Removes all item classes from the collection. """
        self.collection.clear()
    
    def get_all_item_classes(self):
        """ Returns all item classes in the collection.
        
        :returns: All item classes in the collection.
        """
        return self.collection
    
    #--- managing collections --------------------------------------------------
    
    @classmethod
    def get_collection_by_id(cls, collection_id):
        """ Returns item collection with the given id.
        
        :param collection_id: Item class collection ID.
        :returns: Item class collection instance with the given ID.
        """
        if collection_id not in cls.collection_classes:
            raise CollectionDoesNotExist(collection_id)
        return cls.collection_classes[collection_id]



from pprint import pprint


class ModelDeleter(object):
    """ This class keeps track of new items and is able to find those items
    in database that are not new in order to delete or unrefrence them
    (from x-to-many relationships).
    
    :param model_cls: ORM model class used by items that instance of this class
        deals with.
    :param selector_fields: List of fields that are collected from items. The
        values are used later to pull models from database.
    :param keeper_fields: Similar to `selector_fields`. If  a model can be
        pulled from database using values from `keeper_fields`, then this model
        is ignored by this class. 
    """

    def __init__(self, model_cls, selector_fields, keeper_fields):
        self.model_cls = model_cls
        self.selector_fields = selector_fields
        self.keeper_fields = keeper_fields

        self.reset()
    
    def reset(self):
        """ Resets `selectors` and `keepers`. """
        self.selectors = []
        self.keepers = []
    
    def collect_model(self, model):
        """ Saves `selector_fields` from `model`.
        
        :param model: ORM Model to collect field values from.
        """
        filters = self.__collect_filters(model, self.selector_fields)
        if filters not in self.selectors:
            self.selectors.append(filters)
            
        filters = self.__collect_filters(model, self.keeper_fields)
        if filters not in self.keepers:
            self.keepers.append(filters)

    def __collect_filters(self, model, fields):
        filters = {}
        for field in fields:
            if not hasattr(model, field):
                filters[field] = None
            else:
                filters[field] = getattr(model, field)
        return filters
    
    def execute_delete(self, db_adapter):
        """ Deletes rows from database that can be selected (pulled from
        database) using values from `selectors` but excluding those that can
        be selected by `keepers`.
        
        :param db_adapter: Database adapter.
        """
        db_adapter.execute_delete(self.model_cls,
                                  self.selectors, self.keepers)
        self.reset()
    
    def execute_unref(self, parent_model, fkey, db_adapter):
        """ Removes x-to-many references in database from x-to-many
        relationship from `parent_model` to child model through `fkey`
        if refrenced models can be selected using `selectors` values,
        exluding those that can be selected by `keepers`.
        
        :param parent_model: Parent ORM model.
        :param fkey: Foreign key field of `parent_model`.
        :param db_adapter: Database adapter.
        """
        db_adapter.execute_unref(parent_model, fkey,
                                 self.selectors, self.keepers)
        self.reset()
    
    def pprint(self):
        """ Pretty-prints an instance data to console. """
        pprint({
            'selectors': self.selectors,
            'keepers': self.keepers,
            'selector_fields': self.selector_fields,
            'keeper_fields': self.keeper_fields,
        })
        
        